#include <fstream>
#include <stdexcept>
#include <floatSliderPlusImpl.Qt4.h>
#include <qglviewer.h>
#include <QVBoxLayout>
#include <QCheckBox>
#include <assert.h>
#include <QComboBox>
#include <QFileDialog>
#include <QPushButton>
#include <QCheckBox>
#include <QSpinBox>
#include <DataMat.h>
#include "EXR_IO.h"
#include "ModeEqWidget.h"

using namespace std ;
using namespace qglviewer ;

class MyViewer:public QGLViewer
{
	public:
		MyViewer(QWidget *parent,int w,int h,const float *pix)
			: _w(w),_h(h),_pix(pix)
		{
			resize(w,h) ;
			_gain = 1.0 ;
		}

		virtual void draw()
		{
			if(_pix != NULL)
			{
				glPixelTransferf(GL_RED_SCALE,_gain) ;
				glPixelTransferf(GL_GREEN_SCALE,_gain) ;
				glPixelTransferf(GL_BLUE_SCALE,_gain) ;
				glDrawPixels(_w,_h,GL_RGBA,GL_FLOAT,_pix) ;
			}

			glFlush() ;
		}
		void setGain(float f) { _gain = f ; }

	private:
		int _w,_h ;
		float _gain ;
		const float *_pix ;
};

static void checkSize(const DataMat& X,const DataMat& Y)
{
	if(X.NCols()!=Y.NCols()) throw runtime_error("incompatible sizes") ;
	if(X.NRows()!=Y.NRows()) throw runtime_error("incompatible sizes") ;
}
float *getData(const DataMat& mat,float gamma=1.0)
{
	float *res = new float[mat.NRows()*mat.NCols()] ;
	int nC(mat.NCols()) ;
	int nR(mat.NRows()) ;

	for(int r=0;r<nR;++r)
		for(int c=0;c<nC;++c)
		{
			res[c+nC*(nR-r-1)] = pow(mat(r,c),1.0/gamma) ;
			assert(!isnan(res[c+nC*(nR-r-1)])) ;
		}

	return res ;
}
ModeEqWidget::ModeEqWidget(const string& path)
{
	setupUi(this) ;

	// read eigen file
	//
	cout << "Reading images." << endl ;

	int W=0,H=0 ;
	_eigen_vectors.clear() ;
	_ranges.clear() ;
	_av_policy = false ;
	_gamma = 2.2 ;
	_show_what = 0 ;
	cout << "Reading RbyL.bin" << endl ;
	DataMat R(path+"/RbyL.bin") ;
	cout << "Reading GbyL.bin" << endl ;
	DataMat G(path+"/GbyL.bin") ;
	cout << "Reading BbyL.bin" << endl ;
	DataMat B(path+"/BbyL.bin") ; 

	checkSize(R,G) ;
	checkSize(R,B) ;

	int nR(R.NRows()) ;
	int nC(R.NCols()) ;

	_w = nC ;
	_h = nR ;

	_col_channels.push_back(getData(R,_gamma)) ;
	_col_channels.push_back(getData(G,_gamma)) ;
	_col_channels.push_back(getData(B,_gamma)) ;

	for(int i=0;;++i)
	{
		int w,h ;

		stringstream ss;
		ss << path << "LumOut_" << i << ".bin" ;
		DataMat Lum(ss.str()) ;

		if(Lum.NCols() != nC || Lum.NRows() != nR)
			break ;

		cout << "Read " << ss.str() << endl ;
		checkSize(R,Lum) ;
		_eigen_vectors.push_back(getData(Lum)) ;
	}
//	for (int r(0); r<nR; r++) for(int c(0); c<nC; c++) L(r,c) = max(exp(L(r,c))-0.1, 0.0001) ;

	_means.resize(_eigen_vectors.size(),0.0) ;

	int n= _w*_h ;
	for(uint i=0;i<_eigen_vectors.size();++i)
	{
		for(int j=0;j<n;++j)
			_means[i] += _eigen_vectors[i][j] ;

		_means[i] /= n ;
	}
	_ranges.resize(_eigen_vectors.size(),std::pair<float,float>(-2,10)) ;
//	_coefs.resize(_eigen_vectors.size()) ;

	// build sliders
	
	cout << "Building interface." << endl ;
	_sliders.clear() ;
	QVBoxLayout *layout = new QVBoxLayout(sliderFrame) ;

	QPushButton *bt = new QPushButton("Save EXR image") ;
	QCheckBox *cb = new QCheckBox("Scale around mean") ;
	QComboBox *cb2 = new QComboBox ;
	cb2->addItem("Show equalized image") ;
	cb2->addItem("Show Luminance only") ;
	cb2->addItem("Show chrominance only") ;
	layout->addWidget(bt) ;
	layout->addWidget(cb) ;
	layout->addWidget(cb2) ;

	QObject::connect(bt,SIGNAL(clicked()),this,SLOT(saveExrImage())) ;
	QObject::connect(cb,SIGNAL(toggled(bool)),this,SLOT(setAveragePolicy(bool))) ;
	QObject::connect(cb2,SIGNAL(activated(int)),this,SLOT(setShowWhat(int))) ;

	QSpinBox *sp = new QSpinBox ;
	sp->setMinimum(-1) ;
	sp->setMaximum(_eigen_vectors.size()-1) ;
	sp->setValue(-1) ;

	layout->addWidget(sp) ;

	_use_sse_CB = new QCheckBox("Use sse instructions") ;
	_use_sse_CB->setChecked(true) ;
	layout->addWidget(_use_sse_CB) ;

	FloatSliderPlus *sl = new FloatSliderPlus ;
	sl->setProperty("min",0.0) ;
	sl->setProperty("max",10.0) ;
	sl->setProperty("label",QString("gain")) ;
	sl->setValue(1.0) ;
	layout->addWidget(sl) ;

	QObject::connect(sl,SIGNAL(valueChanged(float)),this,SLOT(changeGain(float))) ;

	QObject::connect(sp,SIGNAL(valueChanged(int)),this,SLOT(coefsChanged(int))) ;
	QObject::connect(_use_sse_CB,SIGNAL(toggled(bool)),this,SLOT(updateImage())) ;

	for(uint i=0;i<_eigen_vectors.size();++i)
	{
		FloatSliderPlus *slider = new FloatSliderPlus ;
		slider->setProperty("min",_ranges[i].first) ;
		slider->setProperty("max",_ranges[i].second) ;
		slider->setValue(1.0) ;
		slider->setProperty("label",QString("Mode coefficient ")+QString::number(i)) ;
		_sliders.push_back(slider) ;

		layout->addWidget(slider) ;

		QObject::connect(slider,SIGNAL(valueChanged(float)),this,SLOT( sliderUpdated())) ;
	}
	layout->addStretch() ;//Item(new QSpacerItem(1000, 0, QSizePolicy::Expanding, QSizePolicy::Minimum)) ;

	// build image
	
	_pix = new float[4*_w*_h] ;
	_current_coefs.resize(_sliders.size()) ;

	QVBoxLayout *l2 = new QVBoxLayout(imageFrame) ;
	_viewer = new MyViewer(NULL,_w,_h,_pix) ; 
	l2->addWidget(_viewer) ;
//	l2->setSizeConstraint(QLayout::SetMinimumSize) ;//QSize(_w,_h)) ;
//	l2->addStretch() ;//Item(new QSpacerItem(1000, 0, QSizePolicy::Expanding, QSizePolicy::Minimum)) ;
	adjustSize() ;

	sliderUpdated() ;
}

void ModeEqWidget::changeGain(float v)
{
	_viewer->setGain(v) ;
	updateImage() ;
}
ModeEqWidget::~ModeEqWidget()
{
	for(uint i=0;i<_eigen_vectors.size();++i)
		delete[] _eigen_vectors[i] ;
}

void ModeEqWidget::saveExrImage()
{
	QString name = QFileDialog::getSaveFileName(NULL,"Save exr image.",".","EXR files (*.exr)") ;

	if(name.isNull())
		return ;

	t_EXR_IO<float>::SaveEXR(name.toStdString().c_str(),_w,_h,_pix,4) ;
}

void ModeEqWidget::setAveragePolicy(bool b)
{
	_av_policy = b ;
	updateImage() ;
}


void ModeEqWidget::coefsChanged(int n)
{
	for(uint i=0;i<_sliders.size();++i)
		if(i != n && n > -1)
		{
			_sliders[i]->setValue(0.0) ;
			_current_coefs[i] = 0.0 ;
		}
		else
		{
			_sliders[i]->setValue(1.0) ;
			_current_coefs[i] = 1.0 ;
		}

	updateImage() ;
}

void ModeEqWidget::sliderUpdated()
{
	_current_coefs.resize(_sliders.size()) ;

	for(uint i=0;i<_sliders.size();++i)
		_current_coefs[i] = _sliders[i]->value() ;

	updateImage() ;
}
void ModeEqWidget::setShowWhat(int b)
{
	_show_what = b ;
	updateImage() ;
}

#if defined(__GNUC__) && defined(__SSE__)
//#define USE_SSE_IN_VECTOR3
#include <xmmintrin.h>
#endif

typedef	union
			{
				float v[4];  //!< The 4 vector components
				__m128 m128 ;
			} compound;

const __m128 fast_pow( const __m128& base, const __m128& exponent)
{
    __m128 denom = _mm_mul_ps( exponent, base);
    denom = _mm_sub_ps( exponent, denom);
    denom = _mm_add_ps( base, denom);
    return _mm_mul_ps( base, _mm_rcp_ps(denom));
}
	 
const __m128 fast_exp_ps( const __m128& exponent)
{
#ifdef A_FAIRE
    const __m128 x2 = _mm_mul_ps( exponent, exponent);
    const __m128 term1 =  _mm_add_ps( _mm_set1_ps( 0.496879224f), _mm_mul_ps( _mm_set1_ps(0.190809553f), exponent));
    const __m128 term2 = _mm_add_ps( _mm_set1_ps(1.0f), exponent);
    return  _mm_add_ps( term2,  _mm_mul_ps( x2, term1));
#else
	 static const int n = 7 ;
	 __m128 res = _mm_set1_ps(1.0) ;
	 
	 for(int i=n;i>0;--i)
		 res = _mm_add_ps(_mm_set1_ps(1.0),_mm_mul_ps(exponent,_mm_mul_ps(res,_mm_set1_ps(1.0/i)))) ;

	 return res ;
#endif
}

void ModeEqWidget::updateImage()
{
	if(_use_sse_CB->isChecked())
		updateImage_SSE2() ;
	else
		updateImage_basic() ;
}

void ModeEqWidget::updateImage_SSE2()
{
	int n = _w*_h ;

	bzero(_pix,n*sizeof(float)) ;

//	vector<float> means(_eigen_vectors.size(),0.0) ;
	float f = 1.0/_gamma ;

	for(int j=0;j<n/4;j++)
	{
		compound res ;
		res.m128 = _mm_setzero_ps() ;

		if(_av_policy)
			for(uint i=0;i<_eigen_vectors.size();++i)
			{
				compound vec ;
				vec.m128 = _mm_set_ps( _eigen_vectors[i][4*j],_eigen_vectors[i][4*j+1],_eigen_vectors[i][4*j+2],_eigen_vectors[i][4*j+3]) ;

				res.m128 = _mm_add_ps(
												res.m128,
												_mm_add_ps(	_mm_set1_ps(_means[i]),
																_mm_mul_ps( _mm_set1_ps(_current_coefs[i]),
																				_mm_sub_ps(vec.m128,_mm_set1_ps(_means[i]))))) ;
			}
		else
			for(uint i=0;i<_eigen_vectors.size();++i)
			{
				compound vec ;
				vec.m128 = _mm_set_ps( _eigen_vectors[i][4*j],_eigen_vectors[i][4*j+1],_eigen_vectors[i][4*j+2],_eigen_vectors[i][4*j+3]) ;
				res.m128 = _mm_add_ps(res.m128,_mm_mul_ps(_mm_set1_ps(_current_coefs[i]),vec.m128)) ;
			}

		switch(_show_what)
		{
			case 0: {
						  int kk = j << 2 ;
						  compound v ;
						  v.m128 = _mm_max_ps(_mm_set1_ps(0.0001),_mm_sub_ps(fast_exp_ps(res.m128),_mm_set1_ps(0.1))) ;
						  v.m128 = fast_pow(v.m128,_mm_set1_ps(f)) ;

						  compound mc ;

						  mc.m128 = _mm_mul_ps(_mm_set_ps(_col_channels[0][kk],_col_channels[0][kk+1],_col_channels[0][kk+2],_col_channels[0][kk+3]),v.m128) ;

						  int jj = j << 4 ;
						  _pix[   jj] = mc.v[3] ;
						  _pix[ 4+jj] = mc.v[2] ;
						  _pix[ 8+jj] = mc.v[1] ;
						  _pix[12+jj] = mc.v[0] ;
						  ++jj ;

						  mc.m128 = _mm_mul_ps(_mm_set_ps(_col_channels[1][kk],_col_channels[1][kk+1],_col_channels[1][kk+2],_col_channels[1][kk+3]),v.m128) ;

						  _pix[   jj] = mc.v[3] ;
						  _pix[ 4+jj] = mc.v[2] ;
						  _pix[ 8+jj] = mc.v[1] ;
						  _pix[12+jj] = mc.v[0] ;
						  ++jj ;

						  mc.m128 = _mm_mul_ps(_mm_set_ps(_col_channels[2][kk],_col_channels[2][kk+1],_col_channels[2][kk+2],_col_channels[2][kk+3]),v.m128) ;

						  _pix[   jj] = mc.v[3] ;
						  _pix[ 4+jj] = mc.v[2] ;
						  _pix[ 8+jj] = mc.v[1] ;
						  _pix[12+jj] = mc.v[0] ;
						  ++jj ;

						  _pix[   jj] = 1.0 ;
						  _pix[ 4+jj] = 1.0 ;
						  _pix[ 8+jj] = 1.0 ;
						  _pix[12+jj] = 1.0 ;
						  ++jj ;
					  }
					  break ;
			case 1: {
						  compound v ;
						  v.m128 = fast_pow(_mm_max_ps(_mm_set1_ps(0.0001),_mm_sub_ps(fast_exp_ps(res.m128),_mm_set1_ps(0.1))),_mm_set1_ps(f)) ;

						  int jj = j << 4 ;
						  _pix[   jj] = v.v[3] ;
						  _pix[ 4+jj] = v.v[2] ;
						  _pix[ 8+jj] = v.v[1] ;
						  _pix[12+jj] = v.v[0] ;
						  ++jj ;
						  _pix[   jj] = v.v[3] ;
						  _pix[ 4+jj] = v.v[2] ;
						  _pix[ 8+jj] = v.v[1] ;
						  _pix[12+jj] = v.v[0] ;
						  ++jj ;
						  _pix[   jj] = v.v[3] ;
						  _pix[ 4+jj] = v.v[2] ;
						  _pix[ 8+jj] = v.v[1] ;
						  _pix[12+jj] = v.v[0] ;
						  ++jj ;
						  _pix[   jj] = 1.0 ;
						  _pix[ 4+jj] = 1.0 ;
						  _pix[ 8+jj] = 1.0 ;
						  _pix[12+jj] = 1.0 ;
						  ++jj ;
					  }
					  break ;
			case 2:
					  {
						  int jj = j << 4 ;
						  int kk = j << 2 ;
						  compound v ;

						  v.m128 = fast_pow(_mm_set_ps(_col_channels[0][kk],_col_channels[0][kk+1],_col_channels[0][kk+2],_col_channels[0][kk+3]),_mm_set1_ps(_gamma)) ;
						 
						  _pix[   jj] = v.v[3] ;
						  _pix[ 4+jj] = v.v[2] ;
						  _pix[ 8+jj] = v.v[1] ;
						  _pix[12+jj] = v.v[0] ;
						  ++jj ;

						  v.m128 = fast_pow(_mm_set_ps(_col_channels[1][kk],_col_channels[1][kk+1],_col_channels[1][kk+2],_col_channels[1][kk+3]),_mm_set1_ps(_gamma)) ;
						 
						  _pix[   jj] = v.v[3] ;
						  _pix[ 4+jj] = v.v[2] ;
						  _pix[ 8+jj] = v.v[1] ;
						  _pix[12+jj] = v.v[0] ;
						  ++jj ;

						  v.m128 = fast_pow(_mm_set_ps(_col_channels[2][kk],_col_channels[2][kk+1],_col_channels[2][kk+2],_col_channels[2][kk+3]),_mm_set1_ps(_gamma)) ;
						 
						  _pix[   jj] = v.v[3] ;
						  _pix[ 4+jj] = v.v[2] ;
						  _pix[ 8+jj] = v.v[1] ;
						  _pix[12+jj] = v.v[0] ;
						  ++jj ;

						  _pix[   jj] = 1.0 ;
						  _pix[ 4+jj] = 1.0 ;
						  _pix[ 8+jj] = 1.0 ;
						  _pix[12+jj] = 1.0 ;
					  }
		}
	}

	_viewer->updateGL() ;
}

void ModeEqWidget::updateImage_basic()
{
	int n = _w*_h ;

	bzero(_pix,n*sizeof(float)) ;

//	vector<float> means(_eigen_vectors.size(),0.0) ;
	float f = 1.0/_gamma ;

	int jj=0 ;
	for(int j=0;j<n;++j)
	{
		float res = 0.0 ;

		if(_av_policy)
			for(uint i=0;i<_eigen_vectors.size();++i)
				res += _means[i] + _current_coefs[i] * (_eigen_vectors[i][j] - _means[i]) ;
		else
			for(uint i=0;i<_eigen_vectors.size();++i)
				res += _current_coefs[i] * _eigen_vectors[i][j] ;

		switch(_show_what)
		{
			case 0: {
						  float v = pow(max(0.0001,exp(res)-0.1),f) ;

						  _pix[jj++] = _col_channels[0][j]*v ;
						  _pix[jj++] = _col_channels[1][j]*v ;
						  _pix[jj++] = _col_channels[2][j]*v ;
						  _pix[jj++] = 1.0;
					  }
					  break ;
			case 1: {
						  float v = pow(max(0.0001,exp(res)-0.1),f) ;

						  _pix[jj++] = v ;
						  _pix[jj++] = v ;
						  _pix[jj++] = v ;
						  _pix[jj++] = 1.0;
					  }
					  break ;
			case 2:
					  {
						  _pix[jj++] = pow(_col_channels[0][j],_gamma) ;
						  _pix[jj++] = pow(_col_channels[1][j],_gamma) ;
						  _pix[jj++] = pow(_col_channels[2][j],_gamma) ;
						  _pix[jj++] = 1.0;
					  }
		}
	}

	_viewer->updateGL() ;
}


