#include <NHLSOpt.h>
#include <iostream>
#include <stdlib.h>
#include <fstream>
#include <compcol_double.h>
#include <comprow_double.h>
#include <iotext_double.h>
#include <ilupre_double.h>
#include <icpre_double.h>
#include <cg.h>


#define MYDEBUG(a)
extern string OutputPath ;

namespace
{
        double MinSig(.0001) ;
        double MaxSig(10) ;
        const int KSize(3) ;
        const int CostExp(2) ;
        const double lambda(0.0) ;
        const int maxIter = 195 ;
        const double tolerance= 1e-6 ;
        const bool LOGRECURSION = false ;
}


double WFunc (double Wa, double Wb, double Sigma) 
{
//         Wa = exp(Wa) ; Wb = exp(Wb) ; 
//         cout << "\t\t " << (Wa-Wb) << " " << -(Wa-Wb)*(Wa-Wb)/(2.0*Sigma*Sigma) << " " << exp(-(Wa-Wb)*(Wa-Wb)/(2.0*Sigma*Sigma))  ;        
//         return exp(-(Wa-Wb)*(Wa-Wb)/Sigma) ;
        return exp(-pow(Wa-Wb, CostExp)/Sigma) ;
//         return exp(-pow(exp(Wa)-exp(Wb), CostExp)/Sigma) ;
//            return 0 ;
//         return exp(-(Wa*Wa+Wb*Wb)/(2.0*Sigma*Sigma)) ;
//         return 1 ;
} 


/////////////////// Solve two systems of the form Ax=b, where A is specified
/////////////////// in sparse format using RowIndices, ColIndices, Vals
/////////////////// Sorry, but the library complains if I give it const references
////////////////// x1 and x2 are the output solutions for Min/Max envelopes respectively
//////////////////                      Conjugate Gradient Solution
void CGSolve1(int sz,vector<int>& RowIndices,vector<int>& ColIndices,vector<double>& Vals,VECTOR_double& x,VECTOR_double& b)
{
	Coord_Mat_double A(sz, sz, Vals.size(), &Vals[0], &RowIndices[0], &ColIndices[0]) ;
	CompRow_ILUPreconditioner_double M(A);

	double tol = tolerance ;
	int iter(maxIter) ;

	CG(A, x, b, M, iter, tol)  ;
}

void CGSolve(int sz,     vector<int>& RowIndices1,  vector<int>& ColIndices1, 
                         vector<int>& RowIndices2,  vector<int>& ColIndices2,
                         vector<double>& Vals1,  vector<double>& Vals2,
                         VECTOR_double& x1,  VECTOR_double& b1,
                         VECTOR_double& x2,  VECTOR_double& b2
            )
{
	CGSolve1(sz,RowIndices1,ColIndices1,Vals1,x1,b1) ;
	CGSolve1(sz,RowIndices2,ColIndices2,Vals2,x2,b2) ;
}


/////////////////// Computes the max and min envelopes of Input
/////////////////// Given two sets (minima and maxima), constructs an interpolant for each set 
/////////////////// according to Levin04 's solution for colorization. The extrema are used as constraints (interpolation)
// Min and Max are the envelopes, 
// AdditionalMins and AdditionalMaxs are additional constraint locations
// SmoothnessSigScale allows control of preserving edges (fixed to 1.0 in our results)
void NHLSMinMaxModifExtrema(const DataMat& Input, DataMat& Min, DataMat& Max, 
                            const DataMat& AdditionalMins, const DataMat & AdditionalMaxs, 
                            const DataMat& OrigImage,
                            double SmoothnessSigScale, 
                            int ExtremaKernSize, int ExtremaLeniency,                            
                            bool UseOnlyAdditionalExtrema) 
{
        static int Count(0) ;
        
        bool DebugLog(false) ;                
        int nC(Input.NCols()), nR(Input.NRows()) ;        
        
        double DR = Input.GetDR() ;
        Min.resize(Input) ;
        Max.resize(Input) ;
        vector<pair<int, int> > Mins, Maxs ;        
        OrigImage.StationaryPointsStatistical(Mins, Maxs, 1e-5*DR, ExtremaKernSize,  ExtremaLeniency) ;
        
        double MinimaDensity = Mins.size()/(1.0*nR*nC) ;
        double MaximaDensity = Maxs.size()/(1.0*nR*nC) ;
                
        double MinController(0.0), MaxController(0.0) ;
        if (MinimaDensity < 1.0/2500) MinController =  1.0/ExtremaKernSize ;
        if (MaximaDensity < 1.0/2500) MaxController =  1.0/ExtremaKernSize ;
        
        DataMat MaxMat(AdditionalMaxs) ;
        DataMat MinMat(AdditionalMins); 
        if (!UseOnlyAdditionalExtrema) 
        {
                for (int i(0); i<Mins.size(); i++) MinMat(Mins[i].first, Mins[i].second) = -1 ;
                for (int i(0); i<Maxs.size(); i++) MaxMat(Maxs[i].first, Maxs[i].second) = 1 ;
        }
                
        DataMat ConstraintMat = Input ; 
        int sz(nC*nR) ;  
        vector<int> RowIndices1, ColIndices1 ;
        vector<int> RowIndices2, ColIndices2 ;
        vector<double> Vals1, Vals2; 
                     
        VECTOR_double x1(sz, 0.0), b1(sz, 0.0);
        VECTOR_double x2(sz, 0.0), b2(sz, 0.0);
                
        int nPixels(nC*nR) ;        
        
        DataMat Sig, Sig2, Sig3 ;
        Sig.resize(Input) ; Sig2.resize(Input) ; Sig3.resize(Input) ;
        
        int PrevValIndex1(0), PrevValIndex2(0) ; // Usd for normalization ...
        int SkipCnt (nPixels/1000+2) ;
        char dispch[] = { '-', '/', '|', '\\'};
        unsigned int dispdash(0); 
        for (int ind1(0); ind1<nPixels; ind1++)
        {
               if (!(ind1%SkipCnt)) cout << "\t\t\t Computing Weight Matrix \t\t " << ind1/(nPixels*1.0) << "\t\t\r" ;
        
                bool MinFound(false), MaxFound(false) ;

                ///////////////////////// Check if the current pixel is in the set of stationary points
                int r(ind1/nC), c(ind1%nC) ;                
                if (MinMat(r,c)<0) MinFound=true; 
                if (MaxMat(r,c)>0) MaxFound=true; 
                
                // Enforce minimum extrema density
                double R1(rand()/(1.0*RAND_MAX)), R2(rand()/(1.0*RAND_MAX)) ;
                if (R1<MinController) MinFound=true;                 
                if (R2<MaxController) MaxFound=true; 
                
                double sigma(0) ;   
                if (MinFound) b1(ind1) = Input(r,c) ;
                if (MaxFound) b2(ind1) = Input(r,c) ;
                //////////////////////////////// Compute local mean/variance 
                if (!MinFound || !MaxFound)
                {
                        ///////////////////////// Find Mean, then Find Variance for determining Sigma
                        double Mean(0.0);
                        int KSize2(3);
			if (Count%2) KSize2=5;
                        for (int i(0);i<KSize2; i++) 
                                for (int j(0);j<KSize2; j++)
                                { 
                                        int l=r-KSize2/2+i;
                                        int m=c-KSize2/2+j; 
                                        if (l<0 || l>=nR || m<0 || m>=nC) continue ;
                                        Mean += ConstraintMat(l,m) ;
                                }
                        Mean/=(KSize2*KSize2);
                        double RefVal(ConstraintMat(r,c)), MinDev(0.0), Var(0.0) ;
                        for (int i(0);i<KSize2; i++) 
                                for (int j(0);j<KSize2; j++)
                                { 
                                        int l=r-KSize2/2+i;
                                        int m=c-KSize2/2+j;                                         
                                        if (l<0 || l>=nR || m<0 || m>=nC) continue ;
                                        Var+= pow((ConstraintMat(l,m)-Mean),2) ;                                 
                                        MinDev = max(MinDev, pow((ConstraintMat(l,m)-RefVal),2)) ; 
                                }
                        Var /= (KSize2*KSize2);                                          
                        
                        sigma = (max(0.6*Var, -MinDev/log(0.01))) ;                                
                        sigma = min(max(MinSig, sigma), MaxSig) ;
                        Sig(r,c) = sigma ;       
			///////////////////////// Populate Matrix A. Build row and column indices
                        double Normalizn(0.0) ;
                        double NMax(0.0), NMin(0.0) ;
                        for (int i(0);i<KSize; i++)
                        {                        
                                for (int j(0);j<KSize; j++)
                                {       
                                        int l=r-KSize/2+i;
                                        int m=c-KSize/2+j;
                                        if (l==r && m==c) continue ;
                                        if (l<0 || l>=nR || m<0 || m>=nC) continue ;
                                        int ind2(l*nC+m) ;       
                                        
                                        double Weight = -WFunc(SmoothnessSigScale * ConstraintMat(r, c), SmoothnessSigScale * ConstraintMat(l,m), sigma) ;
                                        double WeightMax, WeightMin ;
                                        WeightMax = WeightMin = Weight ;
                                        Normalizn += Weight ;                                                                           
                                        NMax += WeightMax ;
                                        NMin += WeightMin ;
                                        if (!MinFound)
                                        {
                                                Vals1.push_back(WeightMin) ; 
                                                RowIndices1.push_back(ind1) ; ColIndices1.push_back(ind2) ;                                         
                                        }
                                        if(!MaxFound)
                                        {
                                                Vals2.push_back(WeightMax) ; 
                                                RowIndices2.push_back(ind1) ; ColIndices2.push_back(ind2) ;                                         
                                        }
                                }
                        }
                        if (fabs(Normalizn)>1.e-25) Normalizn = fabs(1.0/Normalizn) ; else {Normalizn = 1.0 ; /*; cout << "!" << endl;*/}
                        if (fabs(NMin)>1.e-25) NMin = fabs(1.0/NMin) ; else {NMin= 1.0 ;}
                        if (fabs(NMax)>1.e-25) NMax= fabs(1.0/NMax) ; else {NMax = 1.0 ;}
                        if (!MinFound) for(int nn(PrevValIndex1); nn<Vals1.size(); nn++) Vals1[nn] *= NMin;
                        if (!MaxFound) for(int nn(PrevValIndex2); nn<Vals2.size(); nn++) Vals2[nn] *= NMax;
                }
                        
                // Diagonal set to 1
                Vals1.push_back(1.0) ; RowIndices1.push_back(ind1) ; ColIndices1.push_back(ind1) ;
                Vals2.push_back(1.0) ; RowIndices2.push_back(ind1) ; ColIndices2.push_back(ind1) ;
                
                PrevValIndex1 = Vals1.size() ;
                PrevValIndex2 = Vals2.size() ;                                
        }
        

	//////////////////////////////// Solve (using ILU preconditioner + Conjugate Gradient)
              CGSolve(sz, RowIndices1, ColIndices1,  RowIndices2,  ColIndices2, Vals1, Vals2, x1,  b1, x2,  b2 ) ;	  
        
       ////////////////////////////////////////////////// Clamp to Input
        for(int i(0); i<nR; i++) 
                for(int j(0); j<nC; j++) 
        {
                Min(i,j) = min(x1[i*nC+j], Input(i,j)) ;       
                Max(i,j) =  max(x2[i*nC+j], Input(i,j)) ;
        }        
        cout << endl ;
        Count++ ;
}

